"""
Copyright 2016 Stephen Boyd, Enzo Busseti, Steven Diamond, BlackRock Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import pickle

import cvxpy as cvx
import numpy as np
import pandas as pd
import copy

from cvxportfolio import simulator, HcostModel, TcostModel, SinglePeriodOpt
from cvxportfolio import ReturnsForecast, MultipleReturnsForecasts, FullSigma
from .base_test import BaseTest

DATAFILE = os.path.dirname(__file__) + os.path.sep + 'sample_data.pickle'


class TestWhatIf(BaseTest):

    def setUp(self):
        with open(DATAFILE, 'rb') as f:
            data = pickle.load(f)
        self.returns, self.sigma, self.volume, self.a, \
            self.b, self.s = data
        self.universe = self.returns.columns
        self.times = self.returns.index

    def test_attribution(self):
        """Test attribution.
        """
        # Alpha source
        alpha_sources = [ReturnsForecast(
            self.returns, name=i) for i in range(3)]
        weights = np.array([0.1, 0.3, 0.6])
        alpha_model = MultipleReturnsForecasts(alpha_sources, weights)
        emp_Sigma = np.cov(self.returns.as_matrix().T)
        risk_model = FullSigma(emp_Sigma, gamma=100.)
        tcost_model = TcostModel(self.volume, self.sigma, self.a, self.b)
        hcost_model = HcostModel(self.s, self.s*0)
        pol = SinglePeriodOpt(alpha_model, [risk_model, tcost_model,
                                            hcost_model], [],
                              solver=cvx.ECOS)

        tcost = TcostModel(self.a, self.b, self.sigma, self.volume)
        hcost = HcostModel(self.s)
        market_sim = simulator.MarketSimulator(self.returns,
                                               costs=[tcost, hcost],
                                               market_volumes=self.volume)

        p_0 = pd.Series(index=self.universe, data=1E6)
        noisy = market_sim.run_backtest(p_0, self.returns.index[1],
                                        self.returns.index[10], pol)
        # linear fit attribution
        attr = market_sim.attribute(noisy, pol,
                                    parallel=False, fit="linear")
        base_line = noisy.v - sum(p_0)
        for i in range(3):
            self.assertItemsAlmostEqual(
                attr[i]/weights[i]/sum(p_0), base_line/sum(p_0))
        self.assertItemsAlmostEqual(attr['RMS error'], np.zeros(len(noisy.v)))

        # least-squares fit attribution
        attr = market_sim.attribute(noisy, pol,
                                    parallel=False, fit="least-squares")
        base_line = noisy.v - sum(p_0)
        for i in range(3):
            self.assertItemsAlmostEqual(
                attr[i]/weights[i]/sum(p_0), base_line/sum(p_0))
        # Residual always 0.
        alpha_sources = [ReturnsForecast(
            self.returns*0, name=i) for i in range(3)]
        weights = np.array([0.1, 0.3, 0.6])
        alpha_model = MultipleReturnsForecasts(alpha_sources, weights)
        pol = copy.copy(pol)
        pol.alpha_model = alpha_model
        attr = market_sim.attribute(noisy, pol,
                                    parallel=False, fit="least-squares")
        self.assertItemsAlmostEqual(attr['residual'], np.zeros(len(noisy.v)))

    def test_attribute_non_profit_series(self):
        """Test attributing series quantities besides profit.
        """
        # Alpha source
        alpha_sources = [ReturnsForecast(
            self.returns, name=i) for i in range(3)]
        weights = np.array([0.1, 0.3, 0.6])
        alpha_model = MultipleReturnsForecasts(alpha_sources, weights)
        emp_Sigma = np.cov(self.returns.as_matrix().T)
        risk_model = FullSigma(emp_Sigma, gamma=100.)
        tcost_model = TcostModel(self.a, self.b, self.sigma, self.volume)
        hcost_model = HcostModel(self.s, self.s*0)
        pol = SinglePeriodOpt(alpha_model, [risk_model, tcost_model,
                                            hcost_model], [],
                              solver=cvx.ECOS)

        tcost = TcostModel(self.volume, self.sigma, self.a, self.b)
        hcost = HcostModel(self.s)
        market_sim = simulator.MarketSimulator(self.returns,
                                               costs=[tcost, hcost],
                                               market_volumes=self.volume)

        p_0 = pd.Series(index=self.universe, data=1E6)
        noisy = market_sim.run_backtest(p_0, self.returns.index[1],
                                        self.returns.index[10], pol)
        # Select tcosts.

        def selector(result):
            return result.leverage

        # linear fit attribution
        attr = market_sim.attribute(noisy, pol, selector,
                                    parallel=False, fit="linear")
        base_line = noisy.leverage
        for i in range(3):
            self.assertItemsAlmostEqual(
                attr[i]/weights[i]/sum(p_0), base_line/sum(p_0))
        self.assertItemsAlmostEqual(attr['RMS error'], np.zeros(len(noisy.v)))

        # least-squares fit attribution
        attr = market_sim.attribute(noisy, pol, selector,
                                    parallel=False, fit="least-squares")
        for i in range(3):
            self.assertItemsAlmostEqual(
                attr[i]/weights[i]/sum(p_0), base_line/sum(p_0))
        # Residual always 0.
        alpha_sources = [ReturnsForecast(
            self.returns*0, name=i) for i in range(3)]
        weights = np.array([0.1, 0.3, 0.6])
        alpha_model = MultipleReturnsForecasts(alpha_sources, weights)
        pol = copy.copy(pol)
        pol.alpha_model = alpha_model
        attr = market_sim.attribute(noisy, pol, selector,
                                    parallel=False, fit="least-squares")
        self.assertItemsAlmostEqual(attr['residual'], np.zeros(len(noisy.v)))

    def test_attribute_non_profit_scalar(self):
        """Test attributing scalar quantities besides profit.
        """
        # Alpha source
        alpha_sources = [ReturnsForecast(
            self.returns, name=i) for i in range(3)]
        weights = np.array([0.1, 0.3, 0.6])
        alpha_model = MultipleReturnsForecasts(alpha_sources, weights)
        emp_Sigma = np.cov(self.returns.as_matrix().T)
        risk_model = FullSigma(emp_Sigma)
        tcost_model = TcostModel(self.a, self.b, self.sigma, self.volume)
        hcost_model = HcostModel(self.s)
        pol = SinglePeriodOpt(
            alpha_model, [100*risk_model, tcost_model, hcost_model], [])

        market_sim = simulator.MarketSimulator(self.returns,
                                               costs=[tcost_model, hcost_model]
                                               )

        p_0 = pd.Series(index=self.universe, data=1E6)
        noisy = market_sim.run_backtest(p_0, self.returns.index[1],
                                        self.returns.index[10], pol)
        # Select tcosts.

        def selector(result):
            return pd.Series(index=[noisy.h.index[-1]],
                             data=result.volatility)

        # linear fit attribution
        attr = market_sim.attribute(noisy, pol, selector,
                                    parallel=False, fit="linear")
        base_line = noisy.volatility
        for i in range(3):
            self.assertAlmostEqual(
                attr[i][0]/weights[i]/sum(p_0), base_line/sum(p_0))
        self.assertItemsAlmostEqual(attr['RMS error'], np.zeros(len(noisy.v)))

        # least-squares fit attribution
        attr = market_sim.attribute(noisy, pol, selector,
                                    parallel=False, fit="least-squares")
        for i in range(3):
            self.assertAlmostEqual(
                attr[i][0]/weights[i]/sum(p_0), base_line/sum(p_0))
        # Residual always 0.
        alpha_sources = [ReturnsForecast(
            self.returns*0, name=i) for i in range(3)]
        weights = np.array([0.1, 0.3, 0.6])
        alpha_model = MultipleReturnsForecasts(alpha_sources, weights)
        pol = copy.copy(pol)
        pol.alpha_model = alpha_model
        attr = market_sim.attribute(noisy, pol, selector,
                                    parallel=False, fit="least-squares")
        self.assertItemsAlmostEqual(attr['residual'], np.zeros(len(noisy.v)))



"""
Copyright 2016 Stephen Boyd, Enzo Busseti, Steven Diamond, BlackRock Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


from abc import ABCMeta, abstractmethod
import cvxpy as cvx
import pandas as pd

__all__ = ['LongOnly', 'LeverageLimit', 'LongCash', 'MaxTrade']


class BaseConstraint(object):
    __metaclass__ = ABCMeta

    def __init__(self, **kwargs):
        self.w_bench = kwargs.pop('w_bench', 0.)

    def weight_expr(self, t, w_plus, z, v):
        """Returns a list of trade constraints.

        Args:
          t: time
          w_plus: post-trade weights
          z: trade weights
          v: portfolio value
        """
        if w_plus is None:
            return self._weight_expr(t, None, z, v)
        return self._weight_expr(t, w_plus-self.w_bench, z, v)

    @abstractmethod
    def _weight_expr(self, t, w_plus, z, v):
        pass


class MaxTrade(BaseConstraint):
    """A limit on maximum trading size.
    """
    def __init__(self, ADVs, max_fraction=0.05, **kwargs):
        self.ADVs = ADVs
        self.max_fraction = max_fraction
        super(MaxTrade, self).__init__(**kwargs)

    def _weight_expr(self, t, w_plus, z, v):
        """Returns a list of trade constraints.

        Args:
          t: time
          w_plus: post-trade weights
          z: trade weights
          v: portfolio value
        """
        return cvx.abs(z[:-1])*v <= self.ADVs.loc[t].values * self.max_fraction
        # TODO check [:-1] and fix pandas <=


class LongOnly(BaseConstraint):
    """A long only constraint.
    """
    def __init__(self, **kwargs):
        super(LongOnly, self).__init__(**kwargs)

    def _weight_expr(self, t, w_plus, z, v):
        """Returns a list of holding constraints.

        Args:
          t: time
          wplus: holdings
        """
        return w_plus >= 0


class LeverageLimit(BaseConstraint):
    """A limit on leverage.

    Attributes:
      limit: A series or number giving the leverage limit.
    """
    def __init__(self, limit, **kwargs):
        self.limit = limit
        super(LeverageLimit, self).__init__(**kwargs)

    def _weight_expr(self, t, w_plus, z, v):
        """Returns a list of holding constraints.

        Args:
          t: time
          wplus: holdings
        """
        if isinstance(self.limit, pd.Series):
            limit = self.limit.loc[t]
        else:
            limit = self.limit
        return cvx.norm(w_plus, 1) <= limit


class LongCash(BaseConstraint):
    """Requires that cash be non-negative.
    """

    def __init__(self, **kwargs):
        super(LongCash, self).__init__(**kwargs)

    def _weight_expr(self, t, w_plus, z, v):
        """Returns a list of holding constraints.

        Args:
          t: time
          wplus: holdings
        """
        return w_plus[-1] >= 0



"""
Copyright 2016 Stephen Boyd, Enzo Busseti, Steven Diamond, BlackRock Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from abc import ABCMeta, abstractmethod
import pandas as pd
import numpy as np
import logging
import cvxpy as cvx

from .costs import BaseCost
from .returns import BaseReturnsModel
from .constraints import BaseConstraint


__all__ = ['Hold', 'FixedTrade', 'PeriodicRebalance', 'AdaptiveRebalance',
           'SinglePeriodOpt', 'MultiPeriodOpt', 'ProportionalTrade']


class BasePolicy(object):
    """ Base class for a trading policy. """
    __metaclass__ = ABCMeta

    def __init__(self):
        self.costs = []
        self.constraints = []

    @abstractmethod
    def get_trades(self, portfolio, t):
        """Trades list given current portfolio and time t.
        """
        return NotImplemented

    def _nulltrade(self, portfolio):
        return pd.Series(index=portfolio.index, data=0.)


class Hold(BasePolicy):
    """Hold initial portfolio.
    """
    def get_trades(self, portfolio, t):
        return self._nulltrade(portfolio)


class ProportionalTrade(BasePolicy):
    """Gets to target in given time steps."""
    def __init__(self, targetweight, time_steps):
        self.targetweight = targetweight
        self.time_steps = time_steps
        super(ProportionalTrade, self).__init__()

    def get_trades(self, portfolio, t):
        try:
            missing_time_steps = len(
                self.time_steps)-next(i for (i, x)
                                      in enumerate(self.time_steps)
                                      if x == t)
        except StopIteration:
            raise Exception(
                "ProportionalTrade can only trade on the given time steps")
        deviation = self.targetweight-portfolio/sum(portfolio)
        return sum(portfolio)*deviation/missing_time_steps


class SellAll(BasePolicy):
    """Sell all non-cash assets."""
    def get_trades(self, portfolio, t):
        trade = -pd.Series(portfolio, copy=True)
        trade.ix[-1] = 0.
        return trade


class FixedTrade(BasePolicy):
    """Trade a fixed trade vector.
    """
    def __init__(self, tradevec=None, tradeweight=None):
        """Trade the tradevec vector (dollars) or tradeweight weights."""
        if tradevec is not None and tradeweight is not None:
            raise(Exception(
                'only one of tradevec and tradeweight can be passed'))
        if tradevec is None and tradeweight is None:
            raise(Exception('one of tradevec and tradeweight must be passed'))
        self.tradevec = tradevec
        self.tradeweight = tradeweight
        assert(self.tradevec is None or sum(self.tradevec) == 0.)
        assert(self.tradeweight is None or sum(self.tradeweight) == 0.)
        super(FixedTrade, self).__init__()

    def get_trades(self, portfolio, t):
        if self.tradevec is not None:
            return self.tradevec
        return sum(portfolio)*self.tradeweight


class BaseRebalance(BasePolicy):

    def _rebalance(self, portfolio):
        return sum(portfolio) * self.target - portfolio


class PeriodicRebalance(BaseRebalance):
    """Track a target portfolio, rebalancing at given times.
    """
    def __init__(self, target, period, **kwargs):
        """
        Args:
            target: target weights, n+1 vector
            period: supported options are "day", "week", "month", "quarter",
                "year".
                rebalance on the first day of each new period
        """
        self.target = target
        self.period = period
        super(PeriodicRebalance, self).__init__()

    def is_start_period(self, t):
        result = not getattr(t, self.period) == getattr(self.last_t,
                                                        self.period) if \
                                                        hasattr(self,
                                                                'last_t')\
                                                        else True
        self.last_t = t
        return result

    def get_trades(self, portfolio, t):
        return self._rebalance(portfolio) if self.is_start_period(t) else \
            self._nulltrade(portfolio)


class AdaptiveRebalance(BaseRebalance):
    """ Rebalance portfolio when deviates too far from target.
    """
    def __init__(self, target, tracking_error):
        self.target = target
        self.tracking_error = tracking_error
        super(AdaptiveRebalance, self).__init__()

    def get_trades(self, portfolio, t):
        weights = portfolio/sum(portfolio)
        diff = (weights - self.target).values

        if np.linalg.norm(diff, 2) > self.tracking_error:
            return self._rebalance(portfolio)
        else:
            return self._nulltrade(portfolio)


class SinglePeriodOpt(BasePolicy):

    def __init__(self, alpha_model, costs, constraints, solver=None,
                 solver_opts={}):

        self.alpha_model = alpha_model
        solver_opts = solver_opts
        assert isinstance(self.alpha_model, BaseReturnsModel)

        super(SinglePeriodOpt, self).__init__()

        for cost in costs:
            assert isinstance(cost, BaseCost)
            self.costs.append(cost)

        for constraint in constraints:
            assert isinstance(constraint, BaseConstraint)
            self.constraints.append(constraint)

        self.solver = solver
        self.solver_opts = solver_opts

    def get_trades(self, portfolio, t):

        value = sum(portfolio)
        w = portfolio/value
        z = cvx.Variable(w.size)  # TODO pass index
        wplus = w.values + z

        alpha_term = self.alpha_model.weight_expr(t, wplus)
        assert(alpha_term.is_concave())

        costs, constraints = [], []

        for cost in self.costs:
            cost_expr, const_expr = cost.weight_expr(t, wplus, z, value)
            costs.append(cost_expr)
            constraints += const_expr

        constraints += [item for item in (con.weight_expr(t, wplus, z, value)
                                          for con in self.constraints)]

        for el in costs:
            assert (el.is_convex())

        for el in constraints:
            assert (el.is_dcp())

        prob = cvx.Problem(
            cvx.Maximize(alpha_term - sum(costs)),
            [cvx.sum_entries(z) == 0] + constraints)
        try:
            prob.solve(solver=self.solver, **self.solver_opts)

            if prob.status == 'unbounded':
                logging.error(
                    'The problem is unbounded. Defaulting to no trades')
                return self._nulltrade(portfolio)

            if prob.status == 'infeasible':
                logging.error(
                    'The problem is infeasible. Defaulting to no trades')
                return self._nulltrade(portfolio)

            # TODO will have index
            return pd.Series(index=portfolio.index, data=(z.value.A1 * value))
        except cvx.SolverError:
            logging.error(
                'The solver %s failed. Defaulting to no trades' % self.solver)
            return self._nulltrade(portfolio)

# class LookaheadModel():
#     """Returns the planning periods for multi-period.
#     """
#     def __init__(self, trading_times, period_lens):
#         self.trading_times = trading_times
#         self.period_lens = period_lens
#
#     def get_periods(self, t):
#         """Returns planning periods.
#         """
#         periods = []
#         tau = t
#         for length in self.period_lens:
#             incr = length*pd.Timedelta('1 days')
#             periods.append((tau, tau + incr))
#             tau += incr
#         return periods


class MultiPeriodOpt(SinglePeriodOpt):

    def __init__(self, trading_times,
                 terminal_weights, lookahead_periods=None, *args, **kwargs):
        """
        trading_times: list, all times at which get_trades will be called
        lookahead_periods: int or None. if None uses all remaining periods
        """
        # Number of periods to look ahead.
        self.lookahead_periods = lookahead_periods
        self.trading_times = trading_times
        # Should there be a constraint that the final portfolio is the bmark?
        self.terminal_weights = terminal_weights
        super(MultiPeriodOpt, self).__init__(*args, **kwargs)

    def get_trades(self, portfolio, t):

        value = sum(portfolio)
        assert (value > 0.)
        w = cvx.Constant(portfolio.values/value)

        prob_arr = []
        z_vars = []

        # planning_periods = self.lookahead_model.get_periods(t)
        for tau in \
                self.trading_times[self.trading_times.index(t):
                                   self.trading_times.index(t) +
                                   self.lookahead_periods]:
            # delta_t in [pd.Timedelta('%d days' % i) for i in
            # range(self.lookahead_periods)]:

            #            tau = t + delta_t
            z = cvx.Variable(*w.size)
            wplus = w + z
            obj = self.alpha_model.weight_expr_ahead(t, tau, wplus)

            costs, constr = [], []
            for cost in self.costs:
                cost_expr, const_expr = cost.weight_expr_ahead(
                    t, tau, wplus, z, value)
                costs.append(cost_expr)
                constr += const_expr

            obj -= sum(costs)
            constr += [cvx.sum_entries(z) == 0]
            constr += [con.weight_expr(t, wplus, z, value)
                       for con in self.constraints]

            prob = cvx.Problem(cvx.Maximize(obj), constr)
            prob_arr.append(prob)
            z_vars.append(z)
            w = wplus

        # Terminal constraint.
        if self.terminal_weights is not None:
            prob_arr[-1].constraints += [wplus == self.terminal_weights.values]

        sum(prob_arr).solve(solver=self.solver)
        return pd.Series(index=portfolio.index,
                         data=(z_vars[0].value.A1 * value))



"""
Copyright 2016 Stephen Boyd, Enzo Busseti, Steven Diamond, BlackRock Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import cvxpy as cvx
import numpy as np
import copy
from .expression import Expression
from .utils.data_management import null_checker, time_locator

__all__ = ['HcostModel', 'TcostModel']


class BaseCost(Expression):
    def __init__(self):
        self.gamma = 1.  # it is changed by gamma * BaseCost()

    def weight_expr(self, t, w_plus, z, value):
        cost, constr = self._estimate(t, w_plus, z, value)
        return self.gamma * cost, constr

    def weight_expr_ahead(self, t, tau, w_plus, z, value):
        cost, constr = self._estimate_ahead(t, tau, w_plus, z, value)
        return self.gamma * cost, constr

    def __mul__(self, other):
        """Read the gamma parameter as a multiplication."""
        newobj = copy.copy(self)
        newobj.gamma *= other
        return newobj

    def __rmul__(self, other):
        """Read the gamma parameter as a multiplication."""
        return self.__mul__(other)


class HcostModel(BaseCost):
    """A model for holding costs.

    Attributes:
      borrow_costs: A dataframe of borrow costs.
      dividends: A dataframe of dividends.
    """

    def __init__(self, borrow_costs, dividends=0.):
        null_checker(borrow_costs)
        self.borrow_costs = borrow_costs
        null_checker(dividends)
        self.dividends = dividends
        super(HcostModel, self).__init__()

    def _estimate(self, t, w_plus, z, value):
        """Estimate holding costs.

        Args:
          t: time of estimate
          wplus: holdings
          tau: time to estimate (default=t)
        """
        try:
            w_plus = w_plus[w_plus.index != self.cash_key]
            w_plus = w_plus.values
        except AttributeError:
            w_plus = w_plus[:-1]  # TODO fix when cvxpy pandas ready

        try:
            self.expression = cvx.mul_elemwise(
                time_locator(self.borrow_costs, t), cvx.neg(w_plus))
        except TypeError:
            self.expression = cvx.mul_elemwise(time_locator(
                self.borrow_costs, t).values, cvx.neg(w_plus))
        try:
            self.expression -= cvx.mul_elemwise(
                time_locator(self.dividends, t), w_plus)
        except TypeError:
            self.expression -= cvx.mul_elemwise(
                time_locator(self.dividends, t).values, w_plus)

        return cvx.sum_entries(self.expression), []

    def _estimate_ahead(self, t, tau, w_plus, z, value):
        return self._estimate(t, w_plus, z, value)

    def value_expr(self, t, h_plus, u):
        self.last_cost = -np.minimum(0, h_plus.iloc[:-1]) * \
                time_locator(self.borrow_costs, t)
        self.last_cost -= h_plus.iloc[:-1] * \
            time_locator(self.dividends, t)

        return sum(self.last_cost)

    def optimization_log(self, t):
        return self.expression.value.A1

    def simulation_log(self, t):
        return self.last_cost


class TcostModel(BaseCost):
    """A model for transaction costs.

    Attributes:
      volume: A dataframe of volumes.
      sigma: A dataframe of daily volatities.
      half_spread: A dataframe of bid-ask spreads divided by 2.
      nonlin_coeff: A dataframe of coefficients for the nonlinear cost.
      power: The nonlinear tcost power.
    """

    def __init__(self, half_spread, nonlin_coeff=0., sigma=0., volume=1.,
                 power=1.5):
        null_checker(half_spread)
        self.half_spread = half_spread
        null_checker(sigma)
        self.sigma = sigma
        null_checker(volume)
        self.volume = volume
        null_checker(nonlin_coeff)
        self.nonlin_coeff = nonlin_coeff
        null_checker(power)
        self.power = power
        super(TcostModel, self).__init__()

    def _estimate(self, t, w_plus, z, value):
        """Estimate tcosts given trades.

        Args:
          t: time of estimate
          z: trades
          value: portfolio value

        Returns:
          An expression for the tcosts.
        """

        try:
            z = z[z.index != self.cash_key]
            z = z.values
        except AttributeError:
            z = z[:-1]  # TODO fix when cvxpy pandas ready

        constr = []

        second_term = time_locator(self.nonlin_coeff, t) * \
            time_locator(self.sigma, t) * \
            (value / time_locator(self.volume, t))**(self.power - 1)

        # no trade conditions
        if np.isscalar(second_term):
            if np.isnan(second_term):
                constr += [z == 0]
                second_term = 0
        else:  # it is a pd series
            no_trade = second_term.index[second_term.isnull()]
            second_term[no_trade] = 0
            constr += [z[second_term.index.get_loc(tick)]
                       == 0 for tick in no_trade]

        try:
            self.expression = cvx.mul_elemwise(
                time_locator(self.half_spread, t), cvx.abs(z))
        except TypeError:
            self.expression = cvx.mul_elemwise(
                time_locator(self.half_spread, t).values, cvx.abs(z))
        try:
            self.expression += cvx.mul_elemwise(second_term,
                                                cvx.abs(z)**self.power)
        except TypeError:
            self.expression += cvx.mul_elemwise(
                second_term.values, cvx.abs(z)**self.power)

        return cvx.sum_entries(self.expression), constr

    def value_expr(self, t, h_plus, u):

        u_nc = u.iloc[:-1]
        self.tmp_tcosts = np.abs(u_nc)*time_locator(self.half_spread, t) + \
            time_locator(self.nonlin_coeff, t)*time_locator(self.sigma, t) * \
            np.abs(u_nc)**self.power / \
            (time_locator(self.volume, t)**(self.power - 1))

        return self.tmp_tcosts.sum()

    def optimization_log(self, t):
        try:
            return self.expression.value.A1
        except AttributeError:
            return np.nan

    def simulation_log(self, t):
        # TODO find another way
        return self.tmp_tcosts

    def _estimate_ahead(self, t, tau, w_plus, z, value):
        """Returns the estimate at time t of tcost at time tau.
        """
        return self._estimate(t, w_plus, z, value)

    def est_period(self, t, tau_start, tau_end, w_plus, z, value):
        """Returns the estimate at time t of tcost over given period.
        """
        K = (tau_end - tau_start).days
        tcost, constr = self.weight_expr(t, None, z / K, value)
        return tcost * K, constr



"""
Copyright 2016 Stephen Boyd, Enzo Busseti, Steven Diamond, BlackRock Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import cvxpy as cvx
from cvxportfolio.expression import Expression
from .utils.data_management import time_locator, null_checker

__all__ = ['ReturnsForecast', 'MPOReturnsForecast',
           'MultipleReturnsForecasts']


class BaseReturnsModel(Expression):
    pass


class ReturnsForecast(BaseReturnsModel):
    """A single return forecast.

    Attributes:
      alpha_data: A dataframe of return estimates.
      delta_data: A confidence interval around the estimates.
      half_life: Number of days for alpha auto-correlation to halve.
    """

    def __init__(self, returns, delta=0., gamma_decay=None, name=None):
        null_checker(returns)
        self.returns = returns
        null_checker(delta)
        self.delta = delta
        self.gamma_decay = gamma_decay
        self.name = name

    def weight_expr(self, t, wplus, z=None, v=None):
        """Returns the estimated alpha.

        Args:
          t: time estimate is made.
          wplus: An expression for holdings.
          tau: time of alpha being estimated.

        Returns:
          An expression for the alpha.
        """
        try:
            alpha = cvx.mul_elemwise(time_locator(self.returns, t), wplus)
        except TypeError:
            alpha = cvx.mul_elemwise(
                time_locator(self.returns, t).values, wplus)
        try:
            alpha -= cvx.mul_elemwise(time_locator(self.delta,
                                                   t), cvx.abs(wplus))
        except TypeError:
            alpha -= cvx.mul_elemwise(time_locator(self.delta,
                                                   t).values, cvx.abs(wplus))

        return cvx.sum_entries(alpha)

    def weight_expr_ahead(self, t, tau, wplus):
        """Returns the estimate at time t of alpha at time tau.

        Args:
          t: time estimate is made.
          wplus: An expression for holdings.
          tau: time of alpha being estimated.

        Returns:
          An expression for the alpha.
        """

        alpha = self.weight_expr(t, wplus)
        if tau > t and self.gamma_decay is not None:
            alpha *= (tau-t).days**(-self.gamma_decay)
        return alpha


class MPOReturnsForecast(BaseReturnsModel):
    """A single alpha estimateion.

    Attributes:
      alpha_data: A dict of serieses of return estimates.
    """

    def __init__(self, alpha_data):
        self.alpha_data = alpha_data

    def weight_expr_ahead(self, t, tau, wplus):
        """Returns the estimate at time t of alpha at time tau.

        Args:
          t: time estimate is made.
          wplus: An expression for holdings.
          tau: time of alpha being estimated.

        Returns:
          An expression for the alpha.
        """
        return self.alpha_data[(t, tau)].values.T*wplus


class MultipleReturnsForecasts(BaseReturnsModel):
    """A weighted combination of alpha sources.

    Attributes:
      alpha_sources: a list of alpha sources.
      weights: An array of weights for the alpha sources.
    """
    def __init__(self, alpha_sources, weights):
        self.alpha_sources = alpha_sources
        self.weights = weights

    def weight_expr(self, t, wplus, z=None, v=None):
        """Returns the estimated alpha.

        Args:
            t: time estimate is made.
            wplus: An expression for holdings.
            tau: time of alpha being estimated.

        Returns:
          An expression for the alpha.
        """
        alpha = 0
        for idx, source in enumerate(self.alpha_sources):
            alpha += source.weight_expr(t, wplus) * self.weights[idx]
        return alpha

    def weight_expr_ahead(self, t, tau, wplus):
        """Returns the estimate at time t of alpha at time tau.

        Args:
          t: time estimate is made.
          wplus: An expression for holdings.
          tau: time of alpha being estimated.

        Returns:
          An expression for the alpha.
        """
        alpha = 0
        for idx, source in enumerate(self.alpha_sources):
            alpha += source.weight_expr_ahead(t,
                                              tau, wplus) * self.weights[idx]
        return alpha



"""
Copyright 2016 Stephen Boyd, Enzo Busseti, Steven Diamond, BlackRock Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import pickle

import cvxpy as cvx
import numpy as np
import pandas as pd

from ..policies import SinglePeriodOpt
from ..costs import HcostModel, TcostModel
from ..returns import ReturnsForecast
from ..risks import FullSigma
from .base_test import BaseTest

DATAFILE = os.path.dirname(__file__) + os.path.sep + 'sample_data.pickle'


class TestOptimizer(BaseTest):

    def setUp(self):
        with open(DATAFILE, 'rb') as f:
            self.returns, self.sigma, self.volume, self.a, self.b, self.s = \
                pickle.load(f)
        self.s = self.s + 1e-3
        self.universe = self.returns.columns
        self.times = self.returns.index

    def test_single_period_opt(self):
        """Test single period optimizer.
        """
        # Alpha source
        gamma = 100.
        n = len(self.universe)
        alpha_model = ReturnsForecast(self.returns)
        emp_Sigma = np.cov(self.returns.as_matrix().T) + np.eye(n)*1e-3
        risk_model = FullSigma(emp_Sigma)
        tcost_model = TcostModel(0, self.b, self.sigma, self.volume, power=2)
        hcost_model = HcostModel(self.s*0, self.s)
        pol = SinglePeriodOpt(alpha_model,
                              [gamma*risk_model, tcost_model, hcost_model],
                              [], solver=cvx.ECOS)
        t = self.times[1]
        p_0 = pd.Series(index=self.universe, data=1E6)
        z = pol.get_trades(p_0, t)
        self.assertAlmostEqual(z.sum(), 0)
        # Compare with CP calculation.
        h = z + p_0
        rho = self.b*self.sigma.loc[t]*(sum(p_0)/self.volume.loc[t])
        rho = np.hstack([rho, 0])
        A = 2*gamma*emp_Sigma + 2*np.diag(rho)
        s_val = pd.Series(index=self.returns.columns, data=self.s)
        s_val['cash'] = 0.
        b = self.returns.loc[t] + 2*rho*(p_0/sum(p_0)) + s_val
        h0 = np.linalg.solve(A, b)
        offset = np.linalg.solve(A, np.ones(n))
        nu = (1 - h0.sum())/offset.sum()
        hstar = h0 + nu*offset
        self.assertAlmostEqual(hstar.sum(), 1)
        self.assertItemsAlmostEqual(h/sum(p_0), hstar, places=4)

    # def test_multi_period(self):
    #     """Test multiperiod optimizer.
    #     """
    #     # Alpha source
    #     bmark = pd.Series(index=self.universe, data=0.)
    #     bmark.cash=1
    #     gamma = 100.
    #     n = len(self.universe)
    #     alpha_model = ReturnsForecast(self.returns)
    #     emp_Sigma = np.cov(self.returns.as_matrix().T) + np.eye(n)*1e-3
    #     risk_model = FullSigma(emp_Sigma,gamma_half_life=np.inf)
    #     tcost_model = TcostModel(self.volume, self.sigma,
    #                             self.a*0, self.b, power=2)
    #     hcost_model = HcostModel(self.s*0, self.s)
    #     pol = MultiPeriodOpt(list(self.times)[:3], bmark, 2, alpha_model,
    #                         [gamma*risk_model, tcost_model, hcost_model],
    #                         [], solver=cvx.ECOS)
    #
    #     t = self.times[1]
    #     p_0 =pd.Series(index=self.universe, data=1E6)
    #     z = pol.get_trades(p_0, t)
    #     self.assertAlmostEqual(z.sum(), 0)
    #     # Compare with CP calculation. Terminal constraint.
    #     h = z + p_0
    #     rho=self.b*self.sigma.loc[t]*(sum(p_0)/self.volume.loc[t])
    #     rho=np.hstack([rho, 0])
    #     A = 2*gamma*emp_Sigma + 4*np.diag(rho)
    #     s_val = self.s.loc[t]
    #     s_val['cash'] = 0
    #     b = self.returns.loc[t] + 2*rho*(p_0/sum(p_0) + bmark) + s_val
    #     h0 = np.linalg.solve(A, b) + bmark
    #     offset = np.linalg.solve(A, np.ones(n))
    #     nu = (1 - h0.sum())/offset.sum()
    #     hstar = h0 + nu*offset
    #     self.assertAlmostEqual(hstar.sum(), 1)
    #     self.assertItemsAlmostEqual(h/sum(p_0), hstar, places=4)
    #
    #
    #     pol = MultiPeriodOpt(2, alpha_model, [risk_model, tcost_model,
    #                          hcost_model], [], solver=cvx.ECOS,
    #                                 terminal_constr=False)
    #
    #     t = self.times[1]
    #     p_0 = pd.Series(index=self.universe, data=1E6)
    #     z = pol.get_trades(p_0, t)
    #     self.assertAlmostEqual(z.sum(), 0)
    #     # Compare with CP calculation.
    #     h = z + p_0
    #     rho = self.b*self.sigma.loc[t]*(sum(p_0)/self.volume.loc[t])
    #     rho = np.hstack([rho, 0])
    #     D = np.diag(rho)
    #     A = np.bmat([[2*gamma*emp_Sigma + 4*D, -2*D, np.ones((n,1)),
    #                   np.zeros((n,1))],
    #                  [-2*D, 2*gamma*emp_Sigma, np.zeros((n,1)),
    #                   np.ones((n,1))],
    #                  [np.ones((1,n)), np.zeros((1,n+2))],
    #                  [np.zeros((1,n)), np.ones((1, n)), np.zeros((1,2))]])
    #     s_val = self.s.loc[t]
    #     s_val['cash'] = 0
    #     b = self.returns.loc[t] + 2*rho*p_0/sum(p_0) + s_val
    #     b = np.hstack([b, self.returns.loc[t] + s_val, 1, 1])
    #     x = np.linalg.solve(A, b)
    #     w1 = x[:n]
    #     w2 = x[n:2*n]
    #     self.assertAlmostEqual(w1.sum(), 1)
    #     self.assertAlmostEqual(w2.sum(), 1)
    #     self.assertItemsAlmostEqual(h/sum(p_0), w1, places=4)



"""
Copyright 2016 Stephen Boyd, Enzo Busseti, Steven Diamond, BlackRock Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


# Base class for unit tests.
import unittest
import numpy as np


class BaseTest(unittest.TestCase):
    # AssertAlmostEqual for lists.
    def assertItemsAlmostEqual(self, a, b, places=5):
        a = self.mat_to_list(a)
        b = self.mat_to_list(b)
        for i in range(len(a)):
            self.assertAlmostEqual(a[i], b[i], places)

    # Overriden method to assume lower accuracy.
    def assertAlmostEqual(self, a, b, places=5):
        super(BaseTest, self).assertAlmostEqual(a, b, places=places)

    def mat_to_list(self, mat):
        """Convert a numpy matrix to a list.
        """
        if isinstance(mat, (np.matrix, np.ndarray)):
            return np.asarray(mat).flatten('F').tolist()
        else:
            return mat



"""
Copyright 2017 Stephen Boyd, Enzo Busseti, Steven Diamond, BlackRock Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import pickle

import pandas as pd
import numpy as np

from cvxportfolio import time_matrix_locator, time_locator, null_checker
from .base_test import BaseTest

DATAFILE = os.path.dirname(__file__) + os.path.sep + 'sample_data.pickle'


class TestUtils(BaseTest):

    def setUp(self):
        with open(DATAFILE, 'rb') as f:
            self.returns, self.sigma, self.volume, self.a, self.b, self.s = \
                pickle.load(f)
        self.volume['cash'] = np.NaN

    def test_null_checker(self):
        """Test null check."""
        self.assertRaises(ValueError, null_checker, np.nan)
        self.assertIsNone(null_checker(1.))
        self.assertIsNone(null_checker(self.returns))
        self.returns.iloc[0, 0] = np.NaN
        self.assertRaises(ValueError, null_checker, self.returns)

    def test_time_locator(self):
        """Test time locator."""
        t = self.returns.index[10]
        len(self.returns.columns)
        self.assertTrue(np.allclose(self.returns.loc[t],
                                    time_locator(self.returns, t)))
        self.assertTrue(np.allclose(23,
                                    time_locator(23, t)))
        self.assertTrue(np.allclose(self.returns.loc[t],
                                    time_locator(self.returns.loc[t], t)))

    def test_matrix_locator(self):
        """Test matrix locator."""
        index = ['a', 'b', 'c']
        df = pd.DataFrame(index=index, columns=index,
                          data=[[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        pn = pd.Panel({'1': df, '2': df*2})
        self.assertTrue(np.allclose(df.values,
                                    time_matrix_locator(df, t=12)))
        self.assertTrue(np.allclose(df.values*2,
                                    time_matrix_locator(pn, t='2')))



"""
Copyright 2016 Stephen Boyd, Enzo Busseti, Steven Diamond, BlackRock Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import pickle

import cvxpy as cvx
import numpy as np
import pandas as pd

from ..costs import HcostModel, TcostModel
from ..returns import ReturnsForecast, MultipleReturnsForecasts
from ..constraints import (LongOnly, LeverageLimit, LongCash, MaxTrade)
from .base_test import BaseTest

DATAFILE = os.path.dirname(__file__) + os.path.sep + 'sample_data.pickle'


class TestModels(BaseTest):

    def setUp(self):
        with open(DATAFILE, 'rb') as f:
            self.returns, self.sigma, self.volume, self.a, self.b, self.s = \
                pickle.load(f)
        # self.volume['cash']=np.NaN
        self.universe = self.returns.columns
        self.times = self.returns.index

    def test_alpha(self):
        """Test alpha models.
        """
        # Alpha source
        w = cvx.Variable(len(self.universe))
        source = ReturnsForecast(self.returns)
        t = self.times[1]
        alpha = source.weight_expr(t, w)
        w.value = np.ones(len(self.universe))
        self.assertAlmostEqual(alpha.value, self.returns.loc[t].sum())
        # with delta
        source = ReturnsForecast(self.returns, self.returns/10)
        alpha = source.weight_expr(t, w)
        tmp = np.ones(len(self.universe))
        tmp[0] = -1
        w.value = tmp
        value = self.returns.loc[t].sum() - 2*self.returns.loc[t].values[0]
        value -= self.returns.loc[t].sum()/10
        self.assertAlmostEqual(alpha.value, value)

        # alpha stream
        source1 = ReturnsForecast(self.returns)
        source2 = ReturnsForecast(-self.returns)
        stream = MultipleReturnsForecasts([source1, source2], [1, 1])
        alpha = stream.weight_expr(t, w)
        self.assertEqual(alpha.value, 0)

        stream = MultipleReturnsForecasts([source1, source2], [-1, 1])
        alpha = stream.weight_expr(t, w)
        value = self.returns.loc[t].sum()
        w.value = np.ones(len(self.universe))
        self.assertEqual(alpha.value, -2*value)

        # with exp decay
        w = cvx.Variable(len(self.universe))
        source = ReturnsForecast(self.returns, gamma_decay=2)
        t = self.times[1]
        tau = self.times[3]
        diff = (tau - t).days
        w.value = np.ones(len(self.universe))
        alpha_t = source.weight_expr(t, w)
        alpha_tau = source.weight_expr_ahead(t, tau, w)
        decay = diff**(-2)
        self.assertAlmostEqual(alpha_tau.value, decay*alpha_t.value)

    def test_tcost_value_expr(self):
        """Test the value expression of the tcost.
        """
        n = len(self.universe)
        value = 1e6
        model = TcostModel(half_spread=self.a, nonlin_coeff=0.,
                           sigma=self.sigma, volume=self.volume)
        t = self.times[1]
        z = np.arange(n) - n/2
        z_var = cvx.Variable(n)
        z_var.value = z
        tcost, _ = model.weight_expr(t, None, z_var, value)
        u = pd.Series(index=self.returns.columns, data=z_var.value.A1*value)
        value_expr = model.value_expr(t, None, u)
        self.assertAlmostEqual(tcost.value, value_expr/value)

        model = TcostModel(half_spread=0, nonlin_coeff=self.b,
                           sigma=self.sigma, volume=self.volume, power=2)
        tcost, _ = model.weight_expr(t, None, z_var, value)
        self.b * self.sigma.loc[t] * (value / self.volume.loc[t])
        value_expr = model.value_expr(t, None, u)
        self.assertAlmostEqual(tcost.value, value_expr/value)

        model = TcostModel(half_spread=0, nonlin_coeff=self.b,
                           sigma=self.sigma, volume=self.volume, power=1.5)
        tcost, _ = model.weight_expr(t, None, z_var, value)
        self.b * self.sigma.loc[t] * np.sqrt(value / self.volume.loc[t])
        value_expr = model.value_expr(t, None, u)
        self.assertAlmostEqual(tcost.value, value_expr/value)

        model = TcostModel(half_spread=self.a, nonlin_coeff=self.b,
                           sigma=self.sigma, volume=self.volume)
        tcost, _ = model.weight_expr(t, None, z_var, value)
        value_expr = model.value_expr(t, None, u)
        self.assertAlmostEqual(tcost.value, value_expr/value)

        # with tau
        model = TcostModel(half_spread=self.a, nonlin_coeff=self.b,
                           sigma=self.sigma, volume=self.volume)
        tau = self.times[2]
        tcost, _ = model.weight_expr_ahead(t, tau, None, z_var, value)
        value_expr = model.value_expr(t, None, u)
        self.assertAlmostEqual(tcost.value, value_expr/value)

    def test_tcost(self):
        """Test tcost model.
        """
        n = len(self.universe)
        value = 1e6
        model = TcostModel(half_spread=self.a, nonlin_coeff=0.,
                           sigma=self.sigma, volume=self.volume)
        t = self.times[1]
        z = np.arange(n) - n/2
        z_var = cvx.Variable(n)
        z_var.value = z
        tcost, _ = model.weight_expr(t, None, z_var, value)
        est_tcost_lin = sum(np.abs(z[:-1])*self.a)
        self.assertAlmostEqual(tcost.value, est_tcost_lin)

        model = TcostModel(half_spread=0, nonlin_coeff=self.b,
                           sigma=self.sigma, volume=self.volume, power=2)
        tcost, _ = model.weight_expr(t, None, z_var, value)
        coeff = self.b * self.sigma.loc[t] * \
            (value / self.volume.loc[t])
        est_tcost_nonlin = np.square(z[:-1]).dot(coeff.values)
        self.assertAlmostEqual(tcost.value, est_tcost_nonlin)

        model = TcostModel(half_spread=0, nonlin_coeff=self.b,
                           sigma=self.sigma, volume=self.volume, power=1.5)
        tcost, _ = model.weight_expr(t, None, z_var, value)
        coeff = self.b * self.sigma.loc[t] * \
            np.sqrt(value / self.volume.loc[t])
        est_tcost_nonlin = np.power(np.abs(z[:-1]), 1.5).dot(coeff.values)
        self.assertAlmostEqual(tcost.value, est_tcost_nonlin)

        model = TcostModel(half_spread=self.a, nonlin_coeff=self.b,
                           sigma=self.sigma, volume=self.volume)
        tcost, _ = model.weight_expr(t, None, z_var, value)
        self.assertAlmostEqual(tcost.value, est_tcost_nonlin + est_tcost_lin)

        # with tau
        model = TcostModel(half_spread=self.a, nonlin_coeff=self.b,
                           sigma=self.sigma, volume=self.volume)
        tau = self.times[2]
        tcost, _ = model.weight_expr_ahead(t, tau, None, z_var, value)
        self.assertAlmostEqual(tcost.value, est_tcost_nonlin + est_tcost_lin)

        tau = t + 10*pd.Timedelta('1 days')
        tcost_tau, _ = model.est_period(t, t, tau, None, z_var, value)
        tcost_t, _ = model.weight_expr(t, None, z_var / 10, value)
        tcost_t *= 10
        self.assertAlmostEqual(tcost_tau.value, tcost_t.value)

    def test_hcost(self):
        """Test holding cost model.
        """
        div = self.s/2
        n = len(self.universe)
        wplus = cvx.Variable(n)
        wplus.value = np.arange(n) - n/2
        t = self.times[1]
        model = HcostModel(self.s)
        hcost, _ = model.weight_expr(t, wplus, None, None)
        bcost = sum(wplus[:-1].value*self.s)
        self.assertAlmostEqual(hcost.value, bcost)

        model = HcostModel(self.s*0, div)
        hcost, _ = model.weight_expr(t, wplus, None, None)
        divs = (np.sum(wplus[:-1].value.A1*div))
        self.assertAlmostEqual(-hcost.value, divs)

        model = HcostModel(self.s, div)
        hcost, _ = model.weight_expr(t, wplus, None, None)
        self.assertAlmostEqual(hcost.value, bcost - divs)

    def test_hcost_value_expr(self):
        """Test the value expression of the hcost.
        """
        div = self.s/2
        n = len(self.universe)
        wplus = cvx.Variable(n)
        wplus.value = np.arange(n) - n/2
        t = self.times[1]
        model = HcostModel(self.s)
        hcost, _ = model.weight_expr(t, wplus, None, None)

        value = 1000.
        h_plus = pd.Series(index=self.returns.columns,
                           data=wplus.value.A1*1000)
        value_expr = model.value_expr(t, h_plus, None)

        self.assertAlmostEqual(hcost.value, value_expr/value)

        model = HcostModel(self.s*0, div)
        hcost, _ = model.weight_expr(t, wplus, None, None)
        value_expr = model.value_expr(t, h_plus, None)
        self.assertAlmostEqual(-hcost.value, value_expr/value)

        model = HcostModel(self.s, div)
        hcost, _ = model.weight_expr(t, wplus, None, None)
        value_expr = model.value_expr(t, h_plus, None)
        self.assertAlmostEqual(hcost.value, value_expr/value)

    def test_hold_constrs(self):
        """Test holding constraints.
        """
        n = len(self.universe)
        wplus = cvx.Variable(n)
        t = self.times[1]

        # long only
        model = LongOnly()
        cons = model.weight_expr(t, wplus, None, None)
        wplus.value = np.ones(n)
        assert cons.value
        wplus.value = -np.ones(n)
        assert not cons.value

        # long cash
        model = LongCash()
        cons = model.weight_expr(t, wplus, None, None)
        wplus.value = np.ones(n)
        assert cons.value
        tmp = np.ones(n)
        tmp[-1] = -1
        wplus.value = tmp
        assert not cons.value

        # leverage limit
        model = LeverageLimit(2)
        cons = model.weight_expr(t, wplus, None, None)
        wplus.value = np.ones(n)/n
        assert cons.value
        tmp = np.zeros(n)
        tmp[0] = 4
        tmp[-1] = -3
        wplus.value = tmp
        assert not cons.value
        model = LeverageLimit(7)
        cons = model.weight_expr(t, wplus, None, None)
        tmp = np.zeros(n)
        tmp[0] = 4
        tmp[-1] = -3
        wplus.value = tmp
        assert cons.value

        limits = pd.Series(index=self.times, data=2)
        limits.iloc[1] = 7
        model = LeverageLimit(limits)
        cons = model.weight_expr(t, wplus, None, None)
        tmp = np.zeros(n)
        tmp[0] = 4
        tmp[-1] = -3
        wplus.value = tmp
        assert cons.value
        cons = model.weight_expr(self.times[2], wplus, None, None)
        assert not cons.value

    def test_trade_constr(self):
        """Test trading constraints.
        """
        n = len(self.universe)
        z = cvx.Variable(n)
        t = self.times[1]

        # avg daily value limits.
        value = 1e6
        model = MaxTrade(self.volume, max_fraction=.1)
        cons = model.weight_expr(t, None, z, value)
        tmp = np.zeros(n)
        tmp[:-1] = self.volume.loc[t].values / value * 0.05
        z.value = tmp
        assert cons.value
        z.value = -100*z.value  # -100*np.ones(n)
        assert not cons.value



"""
Copyright 2016 Stephen Boyd, Enzo Busseti, Steven Diamond, BlackRock Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import print_function
import numpy as np
import pandas as pd
import copy
from .policies import MultiPeriodOpt


def getFiscalQuarter(dt):
    """Convert a time to a fiscal quarter.
    """
    year = dt.year
    quarter = (dt.month-1) // 3 + 1
    return "Q%i %s" % (quarter, year)


class SimulationResult():
    """A container for the result of a simulation.

    Attributes:
        h_next: A dataframe of holdings over time.
        u: A dataframe of trades over time.
        tcosts: A series of transaction costs over time.
        borrow_costs: A series of borrow costs over time.
    """
    def __init__(self, initial_portfolio, policy, cash_key, simulator,
                 simulation_times=None, PPY=252,
                 timedelta=pd.Timedelta("1 days")):
        """
        Initialize the result object.

        Args:
            initial_portfolio:
            policy:
            simulator:
            simulation_times:
            PPY:
        """
        self.PPY = PPY
        self.timedelta = timedelta
        self.initial_val = sum(initial_portfolio)
        self.initial_portfolio = copy.copy(initial_portfolio)
        self.cash_key = cash_key
        self.simulator = simulator
        self.policy = policy

    def summary(self):
        print(self._summary_string())

    def _summary_string(self):
        res = ''
        res += 'Number of periods: %d' % self.u.shape[0]+'\n'
        res += 'Initial timestamp: %s' % self.h.index[0]+'\n'
        res += 'Final timestamp: %s' % self.h.index[-1]+'\n'
        res += 'Portfolio return: ' + \
            '%.2f%%' % (self.returns.mean()*100*self.PPY)+'\n'
        res += 'Excess return: ' + \
            '%.2f%%' % (self.excess_returns.mean()*100*self.PPY)+'\n'
        res += 'Excess risk: ' + \
            '%.2f%%' % (self.excess_returns.std()*100*np.sqrt(self.PPY))+'\n'
        res += 'Sharpe ratio: '+'%.2f' % self.sharpe_ratio+'\n'
        res += 'Max. drawdown: '+'%.2f%%' % self.max_drawdown+'\n'
        res += 'Turnover: '+'%.2f%%' % (self.turnover.mean()*100*self.PPY)+'\n'
        res += 'Average policy time: %.3f sec' % self.policy_time.mean()+'\n'
        res += 'Average simulator time: %.3f sec' % self.simulation_time.mean()
        return res

    def log_data(self, name, t, entry):
        try:
            getattr(self, name).loc[t] = entry
        except AttributeError:
            setattr(self, name,
                    (pd.Series if np.isscalar(entry) else
                     pd.DataFrame)(index=[t], data=[entry]))

    def log_policy(self, t, exec_time):
        self.log_data("policy_time", t, exec_time)
        # TODO mpo policy requires changes in the optimization_log methods
        if not isinstance(self.policy, MultiPeriodOpt):
            for cost in self.policy.costs:
                self.log_data("policy_"+cost.__class__.__name__,
                              t, cost.optimization_log(t))

    def log_simulation(self, t, u, h_next, risk_free_return, exec_time):
        self.log_data("simulation_time", t, exec_time)
        self.log_data("u", t, u)
        self.log_data("h_next", t, h_next)
        self.log_data("risk_free_returns", t, risk_free_return)
        for cost in self.simulator.costs:
            self.log_data("simulator_"+cost.__class__.__name__,
                          t, cost.simulation_log(t))

    @property
    def h(self):
        """
        Concatenate initial portfolio and h_next dataframe.

        Infers the timestamp of last element by increasing the final timestamp.
        """
        tmp = self.h_next.shift(1)
        tmp.ix[0] = self.initial_portfolio
        tmp.loc[self.h_next.index[-1] + self.timedelta] = self.h_next.ix[-1]
        return tmp

    @property
    def v(self):
        """The value of the portfolio over time.
        """
        return self.h.sum(axis=1)

    @property
    def profit(self):
        """The profit made, in dollars."""
        return self.v[-1] - self.v[0]

    @property
    def w(self):
        """The weights of the portfolio over time."""
        return (self.h.T / self.v).T

    @property
    def leverage(self):
        """Portfolio leverage"""
        return np.abs(self.w).sum(1)

    @property
    def volatility(self):
        """The annualized, realized portfolio volatility."""
        return np.sqrt(self.PPY) * np.std(self.returns)

    @property
    def mean_return(self):
        """The annualized mean portfolio return."""
        return self.PPY * np.mean(self.returns)

    @property
    def returns(self):
        """The returns R_t = (v_{t+1}-v_t)/v_t
        """
        val = self.v
        return pd.Series(data=val.values[1:]/val.values[:-1] - 1,
                         index=val.index[:-1])

    @property
    def growth_rates(self):
        """The growth rate log(v_{t+1}/v_t)"""
        return np.log(self.returns + 1)

    @property
    def annual_growth_rate(self):
        """The annualized growth rate PPY/T \sum_{t=1}^T log(v_{t+1}/v_t)
        """
        return self.growth_rates.sum()*self.PPY/self.growth_rates.size

    @property
    def annual_return(self):
        """The annualized return in percent.
        """
        ret = self.growth_rates
        return self._growth_to_return(ret.mean())

    def _growth_to_return(self, growth):
        """Convert growth to annualized percentage return.
        """
        return 100*(np.exp(self.PPY*growth) - 1)

    def get_quarterly_returns(self, benchmark=None):
        """The annualized returns for each fiscal quarter.
        """
        ret = self.growth_rates
        quarters = ret.groupby(getFiscalQuarter).aggregate(np.mean)
        return self._growth_to_return(quarters)

    def get_best_quarter(self, benchmark=None):
        ret = self.get_quarterly_returns(benchmark)
        return (ret.argmax(), ret.max())

    def get_worst_quarter(self, benchmark=None):
        ret = self.get_quarterly_returns(benchmark)
        return (ret.argmin(), ret.min())

    @property
    def excess_returns(self):
        return self.returns - self.risk_free_returns

    @property
    def sharpe_ratio(self):
        return np.sqrt(self.PPY) * np.mean(self.excess_returns) / \
            np.std(self.excess_returns)

    @property
    def turnover(self):
        """Turnover ||u_t||_1/v_t
        """
        noncash_trades = self.u.drop(self.cash_key, axis=1)
        return np.abs(noncash_trades).sum(axis=1)/self.v

    @property
    def trading_days(self):
        """The fraction of days with nonzero turnover.
        """
        return (self.turnover.values > 0).sum()/self.turnover.size

    @property
    def max_drawdown(self):
        """The maximum peak to trough drawdown in percent.
        """
        val_arr = self.v.values
        max_dd_so_far = 0
        cur_max = val_arr[0]
        for val in val_arr[1:]:
            if val >= cur_max:
                cur_max = val
            elif 100*(cur_max - val)/cur_max > max_dd_so_far:
                max_dd_so_far = 100*(cur_max - val)/cur_max
        return max_dd_so_far



"""
Copyright 2017 Stephen Boyd, Enzo Busseti, Steven Diamond, BlackRock Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import pandas as pd

__all__ = ['time_matrix_locator', 'time_locator', 'null_checker',
           'non_null_data_args']


def null_checker(obj):
    """Check if obj contains NaN."""
    if (isinstance(obj, pd.Panel) or
        isinstance(obj, pd.DataFrame) or
            isinstance(obj, pd.Series)):
        if np.any(pd.isnull(obj)):
            raise ValueError('Data object contains NaN values', obj)
    elif np.isscalar(obj):
        if np.isnan(obj):
            raise ValueError('Data object contains NaN values', obj)
    else:
        raise TypeError('Data object can only be scalar or Pandas.')


def non_null_data_args(f):
    def new_f(*args, **kwds):
        for el in args:
            null_checker(el)
        for el in kwds.values():
            null_checker(el)
        return f(*args, **kwds)
    return new_f


def time_matrix_locator(obj, t):
    """Retrieve a matrix from a time indexed Panel, or a static DataFrame."""
    if isinstance(obj, pd.Panel):
        return obj.iloc[obj.axes[0].get_loc(t, method='pad')]
    elif isinstance(obj, pd.DataFrame):
        return obj
    else:  # obj not pandas
        raise TypeError('Expected Pandas DataFrame or Panel, got:', obj)


def time_locator(obj, t):
    """Retrieve data from a time indexed DF, or a Series or scalar."""
    if isinstance(obj, pd.DataFrame):
        res = obj.iloc[obj.axes[0].get_loc(t, method='pad')]
        return res
    elif isinstance(obj, pd.Series) or np.isscalar(obj):
        return obj
    else:  # obj not pandas
        raise TypeError(
            'Expected Pandas DataFrame, Series, or scalar. Got:', obj)



"""
Copyright 2016 Stephen Boyd, Enzo Busseti, Steven Diamond, BlackRock Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from abc import ABCMeta, abstractmethod


class Expression(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def weight_expr(self, t, w_plus, z, value):
        """Returns the estimate of cost at time t."""
        pass

    def weight_expr_ahead(self, t, tau, w_plus, z, value):
        """Returns the estimate at time t of cost at time tau.
        """
        return self.weight_expr(t, w_plus, z, value)

    def value_expr(self, t, h_plus, u):
        """Returns the expression at time t, using value representation.

        This should be overridden if the term is used in the simulator.
        """
        return sum(h_plus)*self.weight_expr(t, h_plus/sum(h_plus),
                                            u/sum(u), sum(h_plus))



from setuptools import setup

setup(
    name='cvxportfolio',
    version='0.0.2',
    author='Enzo Busseti and Steven Diamond',
    author_email='ebusseti@stanford.edu, stevend2@stanford.edu',
    packages=['cvxportfolio',
              'cvxportfolio.tests',
              'cvxportfolio.utils'],
    package_dir={'cvxportfolio': 'cvxportfolio'},
    package_data={'cvxportfolio': ['tests/sample_data.pickle']},
    url='http://github.com/cvxgrp/cvxportfolio/',
    license='Apache',
    zip_safe=False,
    description='ConVeX Portfolio Optimization and Simulation toolset.',
    install_requires=["pandas",
                      "numpy",
                      "matplotlib",
                      "cvxpy"],
    use_2to3=True,
)



"""
Copyright 2016 Stephen Boyd, Enzo Busseti, Steven Diamond, BlackRock Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from abc import abstractmethod

import cvxpy as cvx
import numpy as np
import pandas as pd

from .costs import BaseCost

__all__ = ['FullSigma', 'EmpSigma', 'SqrtSigma', 'WorstCaseRisk',
           'RobustFactorModelSigma', 'RobustSigma',  'FactorModelSigma']


def locator(obj, t):
    """Picks last element before t."""
    try:
        return obj.iloc[obj.axes[0].get_loc(t, method='pad')]
    except AttributeError:  # obj not pandas
        return obj


class BaseRiskModel(BaseCost):

    def __init__(self, **kwargs):
        self.w_bench = kwargs.pop('w_bench', 0.)
        super(BaseRiskModel, self).__init__()
        self.gamma_half_life = kwargs.pop('gamma_half_life', np.inf)

    def weight_expr(self, t, w_plus, z, value):
        self.expression = self._estimate(t, w_plus - self.w_bench, z, value)
        return self.gamma * self.expression, []

    @abstractmethod
    def _estimate(self, t, w_plus, z, value):
        pass

    def weight_expr_ahead(self, t, tau, w_plus, z, value):
        """Estimate risk model at time tau in the future."""
        if self.gamma_half_life == np.inf:
            gamma_multiplier = 1.
        else:
            decay_factor = 2**(-1/self.gamma_half_life)
            # TODO not dependent on days
            gamma_init = decay_factor**((tau - t).days)
            gamma_multiplier = gamma_init*(1 - decay_factor)/(1 - decay_factor)

        return gamma_multiplier * self.weight_expr(t, w_plus, z, value)[0], []

    def optimization_log(self, t):
        if self.expression.value:
            return self.expression.value
        else:
            return np.NaN


class FullSigma(BaseRiskModel):
    def __init__(self, Sigma, **kwargs):
        """Sigma is either a matrix or a pd.Panel"""
        self.Sigma = Sigma
        try:
            assert(not pd.isnull(Sigma).values.any())
        except AttributeError:
            assert (not pd.isnull(Sigma).any())
        super(FullSigma, self).__init__(**kwargs)

    def _estimate(self, t, wplus, z, value):
        try:
            self.expression = cvx.quad_form(wplus, locator(self.Sigma, t))
        except TypeError:
            self.expression = cvx.quad_form(
                wplus, locator(self.Sigma, t).values)
        return self.expression


class EmpSigma(BaseRiskModel):
    """Empirical Sigma matrix, built looking at *lookback* past returns."""
    def __init__(self, returns, lookback, **kwargs):
        """returns is dataframe, lookback is int"""
        self.returns = returns
        self.lookback = lookback
        assert(not np.any(pd.isnull(returns)))
        super(EmpSigma, self).__init__(**kwargs)

    def _estimate(self, t, wplus, z, value):
        idx = self.returns.index.get_loc(t)
        # TODO make sure pandas + cvxpy works
        R = self.returns.iloc[max(idx-1-self.lookback, 0):idx-1]
        assert (R.shape[0] > 0)
        self.expression = cvx.sum_squares(R.values*wplus)/self.lookback
        return self.expression


class SqrtSigma(BaseRiskModel):
    def __init__(self, sigma_sqrt, **kwargs):
        """returns is dataframe, lookback is int"""
        self.sigma_sqrt = sigma_sqrt
        assert(not np.any(pd.isnull(sigma_sqrt)))
        super(SqrtSigma, self).__init__(**kwargs)

    def _estimate(self, t, wplus, z, value):
        # TODO make sure pandas + cvxpy works
        self.expression = cvx.sum_squares(wplus.T*self.sigma_sqrt.values)
        return self.expression


class FactorModelSigma(BaseRiskModel):
    def __init__(self, exposures, factor_Sigma, idiosync, **kwargs):
        """Each is a pd.Panel (or ) or a vector/matrix"""
        self.exposures = exposures
        assert (not exposures.isnull().values.any())
        self.factor_Sigma = factor_Sigma
        assert (not factor_Sigma.isnull().values.any())
        self.idiosync = idiosync
        assert(not idiosync.isnull().values.any())
        super(FactorModelSigma, self).__init__(**kwargs)

    def _estimate(self, t, wplus, z, value):
        self.expression = cvx.sum_squares(cvx.mul_elemwise(
            np.sqrt(locator(self.idiosync, t).values), wplus)) + \
            cvx.quad_form((wplus.T*locator(self.exposures, t).values.T).T,
                          locator(self.factor_Sigma, t).values)
        return self.expression


class RobustSigma(BaseRiskModel):
    """Implements covariance forecast error risk."""
    def __init__(self, Sigma, epsilon, **kwargs):
        self.Sigma = Sigma  # pd.Panel or matrix
        self.epsilon = epsilon  # pd.Series or scalar
        super(RobustSigma, self).__init__(**kwargs)

    def _estimate(self, t, wplus, z, value):
        self.expression = cvx.quad_form(wplus, locator(self.Sigma, t)) + \
                             locator(self.epsilon, t) * \
                             (cvx.abs(wplus).T * np.diag(locator(
                                self.Sigma, t)))**2

        return self.expression


class RobustFactorModelSigma(BaseRiskModel):
    """Implements covariance forecast error risk."""
    def __init__(self, exposures, factor_Sigma, idiosync, epsilon, **kwargs):
        """Each is a pd.Panel (or ) or a vector/matrix"""
        self.exposures = exposures
        assert (not exposures.isnull().values.any())
        self.factor_Sigma = factor_Sigma
        assert (not factor_Sigma.isnull().values.any())
        self.idiosync = idiosync
        assert(not idiosync.isnull().values.any())
        self.epsilon = epsilon
        super(RobustFactorModelSigma, self).__init__(**kwargs)

    def _estimate(self, t, wplus, z, value):
        F = locator(self.exposures, t)
        f = (wplus.T * F.T).T
        Sigma_F = locator(self.factor_Sigma, t)
        D = locator(self.idiosync, t)
        self.expression = cvx.sum_squares(
            cvx.mul_elemwise(np.sqrt(D), wplus)) + \
            cvx.quad_form(f, Sigma_F) + \
            self.epsilon * (cvx.abs(f).T * np.sqrt(np.diag(Sigma_F)))**2

        return self.expression


class WorstCaseRisk(BaseRiskModel):
    def __init__(self, riskmodels, **kwargs):
        self.riskmodels = riskmodels
        super(WorstCaseRisk, self).__init__(**kwargs)

    def _estimate(self, t, wplus, z, value):
        self.risks = [risk.weight_expr(t, wplus, z, value)
                      for risk in self.riskmodels]
        return cvx.max_elemwise(*self.risks)

    def optimization_log(self, t):
        """Return data to log in the result object."""
        return pd.Series(index=[model.__class__.__name__ for
                                model in self.riskmodels],
                         data=[risk.value[0, 0] for risk in self.risks])



"""
Copyright 2016 Stephen Boyd, Enzo Busseti, Steven Diamond, BlackRock Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import pickle
import copy

import pandas as pd

from cvxportfolio import TcostModel, HcostModel
from .base_test import BaseTest
from cvxportfolio import MarketSimulator, SimulationResult

DATAFILE = os.path.dirname(__file__) + os.path.sep + 'sample_data.pickle'


class TestSimulator(BaseTest):

    def setUp(self):
        with open(DATAFILE, 'rb') as f:
            self.returns, self.sigma, self.volume, self.a, self.b, self.s = \
                pickle.load(f)
        self.portfolio = pd.Series(index=self.returns.columns, data=1E6)
        self.tcost_term = TcostModel(self.a, self.b, self.sigma, self.volume)
        self.hcost_term = HcostModel(self.s)
        self.Simulator = MarketSimulator(self.returns,
                                         costs=[self.tcost_term,
                                                self.hcost_term],
                                         market_volumes=self.volume)

    def test_propag(self):
        """Test propagation of portfolio."""
        t = self.returns.index[1]
        h = copy.copy(self.portfolio)
        results = SimulationResult(initial_portfolio=h, policy=None,
                                   cash_key='cash', simulator=self.Simulator)
        u = pd.Series(index=self.portfolio.index, data=1E4)
        h_next, u = self.Simulator.propagate(h, u=u, t=t)
        results.log_simulation(t=t, u=u, h_next=h_next,
                               risk_free_return=0., exec_time=0)
        self.assertAlmostEquals(
            results.simulator_TcostModel.sum().sum(), 157.604, 3)
        self.assertAlmostEquals(
            results.simulator_HcostModel.sum().sum(), 0., 3)
        self.assertAlmostEqual(sum(h_next), 28906767.251, 3)

    def test_propag_list(self):
        """Test propagation of portfolio, list of trades."""
        t = self.returns.index[1]
        h = copy.copy(self.portfolio)
        results = SimulationResult(initial_portfolio=h, policy=None,
                                   cash_key='cash', simulator=self.Simulator)
        u = pd.Series(index=self.portfolio.index, data=[1E4]*29)
        h_next, u = self.Simulator.propagate(h, u, t=t)
        results.log_simulation(t=t, u=u, h_next=h_next,
                               risk_free_return=0., exec_time=0)
        self.assertAlmostEquals(
            results.simulator_TcostModel.sum().sum(), 157.604, 3)
        self.assertAlmostEquals(
            results.simulator_HcostModel.sum().sum(), 0., 3)
        self.assertAlmostEqual(sum(h_next), 28906767.251, 3)

    def test_propag_neg(self):
        """Test propagation of portfolio, negative trades."""
        t = self.returns.index[1]
        h = copy.copy(self.portfolio)
        results = SimulationResult(initial_portfolio=h, policy=None,
                                   cash_key='cash', simulator=self.Simulator)
        u = pd.Series(index=self.portfolio.index, data=[-1E4]*29)
        h_next, u = self.Simulator.propagate(h, u, t=t)
        results.log_simulation(t=t, u=u, h_next=h_next,
                               risk_free_return=0., exec_time=0)
        self.assertAlmostEquals(
            results.simulator_TcostModel.sum().sum(), 157.604, 3)
        self.assertAlmostEquals(
            results.simulator_HcostModel.sum().sum(), 0., 3)
        self.assertAlmostEqual(sum(h_next), 28908611.931, 3)

    def test_hcost_pos(self):
        """Test hcost function, positive positions."""
        self.hcost_term.borrow_costs += 0
        t = self.returns.index[1]
        h = copy.copy(self.portfolio)
        results = SimulationResult(initial_portfolio=h, policy=None,
                                   cash_key='cash', simulator=self.Simulator)
        u = pd.Series(index=self.portfolio.index, data=1E4)
        h_next, u = self.Simulator.propagate(h, u, t=t)
        results.log_simulation(t=t, u=u, h_next=h_next,
                               risk_free_return=0., exec_time=0)
        self.assertAlmostEquals(results.simulator_HcostModel.sum().sum(), 0.)

    def test_hcost_neg(self):
        """Test hcost function, negative positions."""
        self.hcost_term.borrow_costs += .0001
        t = self.returns.index[1]
        h = copy.copy(self.portfolio)
        results = SimulationResult(initial_portfolio=h, policy=None,
                                   cash_key='cash', simulator=self.Simulator)
        u = pd.Series(index=self.portfolio.index, data=-2E6)
        h_next, u = self.Simulator.propagate(h, u, t=t)
        results.log_simulation(t=t, u=u, h_next=h_next,
                               risk_free_return=0.,
                               exec_time=0)
        self.assertAlmostEquals(
            results.simulator_HcostModel.sum().sum(), 2800.0)



"""
Copyright 2016 Stephen Boyd, Enzo Busseti, Steven Diamond, BlackRock Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import copy
import logging
import time

import multiprocess
import numpy as np
import pandas as pd
import cvxpy as cvx

from .returns import MultipleReturnsForecasts

from .result import SimulationResult
from .costs import BaseCost

# TODO update benchmark weights (?)
# Also could try jitting with numba.


class MarketSimulator():
    logger = None

    def __init__(self, market_returns, costs,
                 market_volumes=None, cash_key='cash'):
        """Provide market returns object and cost objects."""
        self.market_returns = market_returns
        if market_volumes is not None:
            self.market_volumes = market_volumes[
                market_volumes.columns.difference([cash_key])]
        else:
            self.market_volumes = None

        self.costs = costs
        for cost in self.costs:
            assert (isinstance(cost, BaseCost))

        self.cash_key = cash_key

    def propagate(self, h, u, t):
        """Propagates the portfolio forward over time period t, given trades u.

        Args:
            h: pandas Series object describing current portfolio
            u: n vector with the stock trades (not cash)
            t: current time

        Returns:
            h_next: portfolio after returns propagation
            u: trades vector with simulated cash balance
        """
        assert (u.index.equals(h.index))

        if self.market_volumes is not None:
            # don't trade if volume is null
            null_trades = self.market_volumes.columns[
                self.market_volumes.loc[t] == 0]
            if len(null_trades):
                logging.info('No trade condition for stocks %s on %s' %
                             (null_trades, t))
                u.loc[null_trades] = 0.

        hplus = h + u
        costs = [cost.value_expr(t, h_plus=hplus, u=u) for cost in self.costs]
        for cost in costs:
            assert(not pd.isnull(cost))
            assert(not np.isinf(cost))

        u[self.cash_key] = - sum(u[u.index != self.cash_key]) - sum(costs)
        hplus[self.cash_key] = h[self.cash_key] + u[self.cash_key]

        assert (hplus.index.sort_values().equals(
            self.market_returns.columns.sort_values()))
        h_next = self.market_returns.loc[t] * hplus + hplus

        assert (not h_next.isnull().values.any())
        assert (not u.isnull().values.any())
        return h_next, u

    def run_backtest(self, initial_portfolio, start_time, end_time,
                     policy, loglevel=logging.WARNING):
        """Backtest a single policy.
        """
        logging.basicConfig(level=loglevel)

        results = SimulationResult(initial_portfolio=copy.copy(
                                                        initial_portfolio),
                                   policy=policy, cash_key=self.cash_key,
                                   simulator=self)
        h = initial_portfolio

        simulation_times = self.market_returns.index[
                (self.market_returns.index >= start_time) &
                (self.market_returns.index <= end_time)]
        logging.info('Backtest started, from %s to %s' %
                     (simulation_times[0], simulation_times[-1]))

        for t in simulation_times:
            logging.info('Getting trades at time %s' % t)
            start = time.time()
            try:
                u = policy.get_trades(h, t)
            except cvx.SolverError:
                logging.warning(
                    'Solver failed on timestamp %s. Default to no trades.' % t)
                u = pd.Series(index=h.index, data=0.)
            end = time.time()
            assert (not pd.isnull(u).any())
            results.log_policy(t, end-start)

            logging.info('Propagating portfolio at time %s' % t)
            start = time.time()
            h, u = self.propagate(h, u, t)
            end = time.time()
            assert (not h.isnull().values.any())
            results.log_simulation(t=t, u=u, h_next=h,
                                   risk_free_return=self.market_returns.loc[
                                       t, self.cash_key],
                                   exec_time=end-start)

        logging.info('Backtest ended, from %s to %s' %
                     (simulation_times[0], simulation_times[-1]))
        return results

    def run_multiple_backtest(self, initial_portf, start_time,
                              end_time, policies,
                              loglevel=logging.WARNING, parallel=True):
        """Backtest multiple policies.
        """

        def _run_backtest(policy):
            return self.run_backtest(initial_portf, start_time, end_time,
                                     policy, loglevel=loglevel)

        num_workers = min(multiprocess.cpu_count(), len(policies))
        if parallel:
            workers = multiprocess.Pool(num_workers)
            results = workers.map(_run_backtest, policies)
            workers.close()
            return results
        else:
            return list(map(_run_backtest, policies))

    def what_if(self, time, results, alt_policies, parallel=True):
        """Run alternative policies starting from given time.
        """
        # TODO fix
        initial_portf = copy.copy(results.h.loc[time])
        all_times = results.h.index
        alt_results = self.run_multiple_backtest(initial_portf,
                                                 time,
                                                 all_times[-1],
                                                 alt_policies, parallel)
        for idx, alt_result in enumerate(alt_results):
            alt_result.h.loc[time] = results.h.loc[time]
            alt_result.h.sort_index(axis=0, inplace=True)
        return alt_results

    @staticmethod
    def reduce_signal_perturb(initial_weights, delta):
        """Compute matrix of perturbed weights given initial weights."""
        perturb_weights_matrix = \
            np.zeros((len(initial_weights), len(initial_weights)))
        for i in range(len(initial_weights)):
            perturb_weights_matrix[i, :] = initial_weights / \
                (1 - delta * initial_weights[i])
            perturb_weights_matrix[i, i] = (1 - delta) * initial_weights[i]
        return perturb_weights_matrix

    def attribute(self, true_results, policy,
                  selector=None,
                  delta=1,
                  fit="linear",
                  parallel=True):
        """Attributes returns over a period to individual alpha sources.

        Args:
            true_results: observed results.
            policy: the policy that achieved the returns.
                    Alpha model must be a stream.
            selector: A map from SimulationResult to time series.
            delta: the fractional deviation.
            fit: the type of fit to perform.
        Returns:
            A dict of alpha source to return series.
        """
        # Default selector looks at profits.
        if selector is None:
            def selector(result):
                return result.v - sum(result.initial_portfolio)

        alpha_stream = policy.alpha_model
        assert isinstance(alpha_stream, MultipleReturnsForecasts)
        times = true_results.h.index
        weights = alpha_stream.weights
        assert np.sum(weights) == 1
        alpha_sources = alpha_stream.alpha_sources
        num_sources = len(alpha_sources)
        Wmat = self.reduce_signal_perturb(weights, delta)
        perturb_pols = []
        for idx in range(len(alpha_sources)):
            new_pol = copy.copy(policy)
            new_pol.alpha_model = MultipleReturnsForecasts(alpha_sources,
                                                           Wmat[idx, :])
            perturb_pols.append(new_pol)
        # Simulate
        p0 = true_results.initial_portfolio
        alt_results = self.run_multiple_backtest(p0, times[0], times[-1],
                                                 perturb_pols, parallel)
        # Attribute.
        true_arr = selector(true_results).values
        attr_times = selector(true_results).index
        Rmat = np.zeros((num_sources, len(attr_times)))
        for idx, result in enumerate(alt_results):
            Rmat[idx, :] = selector(result).values
        Pmat = cvx.Variable(num_sources, len(attr_times))
        if fit == "linear":
            prob = cvx.Problem(cvx.Minimize(0), [Wmat * Pmat == Rmat])
            prob.solve()
        elif fit == "least-squares":
            error = cvx.sum_squares(Wmat * Pmat - Rmat)
            prob = cvx.Problem(cvx.Minimize(error),
                               [Pmat.T * weights == true_arr])
            prob.solve()
        else:
            raise Exception("Unknown fitting method.")
        # Dict of results.
        wmask = np.tile(weights[:, np.newaxis], (1, len(attr_times))).T
        data = pd.DataFrame(columns=[s.name for s in alpha_sources],
                            index=attr_times,
                            data=Pmat.value.T.A * wmask)
        data['residual'] = true_arr - np.matrix((weights * Pmat).value).A1
        data['RMS error'] = np.matrix(
            cvx.norm(Wmat * Pmat - Rmat, 2, axis=0).value).A1
        data['RMS error'] /= np.sqrt(num_sources)
        return data
